'''
This file contains definitions for all kernels.
'''

import warnings
from cmath import pi
import math
from abc import abstractmethod, ABCMeta
from scipy.sparse import lil_matrix
from scipy.special import jv
import numpy.matlib as matlib
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity, rbf_kernel, \
    linear_kernel, sigmoid_kernel, chi2_kernel, polynomial_kernel, \
    additive_chi2_kernel, laplacian_kernel
from sklearn.decomposition import PCA
from sklearn.gaussian_process.kernels import RationalQuadratic
from sklearn.preprocessing import MinMaxScaler


def gaussian_kernel(x, y, gamma=None):
    '''
    Gaussian kernel. An alias of sklearn.metrics.pairwise.rbf_kernel
    '''
    return rbf_kernel(x, y, gamma)


def exponential_kernel(x, y, gamma=None):
    '''
    We use exponental and laplacian kernels interchangably.
    '''
    return laplacian_kernel(x, y, gamma=gamma)


def anova_kernel(x, y, sigma=1, d=2):
    if sigma is None:
        sigma = 1
    if d is None:
        d = 2
    return ANOVA(sigma=sigma, d=d)(x, y)


def rq_kernel(x, y, a=1, l=1):
    '''
    The Rational Quadratic kernel. Using sklearn.gaussian_process.kernels.RationalQuadratic.
    k(x_i, x_j) = \\left(1 + \\frac{d(x_i, x_j)^2 }{ 2\\alpha  l^2}\\right)^{-\\alpha}

    Paramters
    ---------
    a - scale mixture
    l - length scale    
    '''
    return RationalQuadratic(length_scale=l, alpha=a)(x, y)


def rq_kernel_v2(x, y, c=1):
    '''
    k(x, y) = 1 - ||x-y||^2/(||x-y||^2+c)    
    '''
    return RationalQuadratic2(c=c)(x, y)


def imq_kernel(x, y, c=1):
    return InverseMultiquadratic(c=c)(x, y)


def cauchy_kernel(x, y, sigma=None):  # = img_kernel squared
    return Cauchy(sigma=sigma)(x, y)


def ts_kernel(x, y, d=3):  # d = 2 becomes Cauchy
    return TStudent(degree=d)(x, y)


def spline_kernel(x, y, k=10, flavor=1):
    '''
    The spline kernel. We provide two flavors of implementation.

    Parameters
    ----------
    k : preprocessed dims. 
        The spline kernel doesnot work well on high-dimensional spectroscopic data, 
        so we can perform a pre-processing DimRed.  
    flavor : 1 - self implmentation
        2 - use the Spline() class
    '''

    if k is not None:
        k = min(k, x.shape[1])
        x = PCA(k).fit_transform(x)

    if flavor == 2:
        return Spline()(x, y)

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            prod = 1
            for a, b in zip(x1, x2):
                prod = prod * (1 + a*b + a*b*min(a, b) - (a+b) /
                               2 * min(a, b)**2 + min(a, b)**3 / 3)
            M[idx1, idx2] = prod
    return M


def sorensen_kernel(x, y):
    return Sorensen()(x, y)


def tanimoto_kernel(x, y):
    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            p = np.dot(x1, x2)
            M[idx1, idx2] = p / (np.linalg.norm(x1)**2 +
                                 np.linalg.norm(x2)**2 - p)
    return M


def min_kernel(x, y):
    return Min()(x, y)


def minmax_kernel(x, y):
    return MinMax()(x, y)


def expmin_kernel(x, y, a=1):
    '''
    exponential min kernel.
    K(x,y) = exp(-a min (|x-y|,|x+y|))^2
    '''
    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            # print(np.linalg.norm(x1-x2) , np.linalg.norm(x1+x2))
            M[idx1, idx2] = math.exp(-a * (min(np.linalg.norm(x1-x2),
                                     np.linalg.norm(x1+x2)) ** 2))

    return M


def ghi_kernel(x, y, alpha=1):
    '''
    Generalized Histogram Intersection kernel
    '''
    return GeneralizedHistogramIntersection(alpha=alpha)(x, y)


def fourier_kernel(x, y, q=0.1):

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            d = np.linalg.norm(x1-x2)
            M[idx1, idx2] = (1-q**2) / (2*(1-2*q*math.cos(d)+q**2))
    return M


def fourier_kernel_v2(x, y, q=0.1):
    return Fourier(q=q)(x, y)


def wavelet_kernel(x, y):
    '''
    The wavelet is a family / series. This is just a commonly used specific implementation.
    Use "import pywt, pywt.wavelist()" to check a full list of wavelets.  
    '''
    return Wavelet()(x, y)


def log_kernel(x, y, d=2):
    '''
    On real data, we found the log kernel is insensative to param d. In most cases, use d's default value.
    '''
    return Log(d=d)(x, y)


def power_kernel(x, y, d=2):
    if d is None:
        d = 2
    return Power(d=d)(x, y)


def bessel_kernel(x, y, v=0, s=1):
    # / euclidean_dist_matrix(x,y)
    return jv(v+1, -s * euclidean_dist_matrix(x, y))


def matern_kernel(x, y, v=0.5, s=1):
    '''
    The matern kernel. 
    Implemented according to the math definition.

    Parameter
    ---------
    v : controls smoothness. when v = 1/2, it becomes into the laplacian/exp kernel. 
    s : controls scaling
    '''
    if v is None:
        v = 0.5
    z = math.sqrt(2*v) / s * euclidean_dist_matrix(x, y)
    return 1/(math.gamma(v) * 2**(v-1)) * (z**v) * mod_bessel(z)


def mod_bessel(x):
    return np.sqrt(math.pi / (2*x)) * np.exp(-x)


def ess_kernel(x, y, p=1, s=1):
    '''
    Parameter
    ---------
    p : periodical parameter
    s : scale parameter
    '''
    if p is None:
        p = 1
    if s is None:
        s = 1

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            M[idx1, idx2] = math.exp(-2*math.sin(pi *
                                     np.linalg.norm(x1-x2)/p)/(s**2))

    return M


def fejer_kernel(x, y, k=10):
    '''
    Parameter
    ---------
    k - order of fejer series. Usually we don't use k = 1 as it always equals 1.
    '''
    if k is None:
        k = 10

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            if math.cos(np.linalg.norm(x1-x2)) == 1:
                M[idx1, idx2] = k  # handle divided-by-0 error
            else:
                M[idx1, idx2] = (1-math.cos(k * np.linalg.norm(x1-x2))) / \
                    (1-math.cos(np.linalg.norm(x1-x2))) / k
    return M
    # return ( 1-np.cos(k*euclidean_dist_matrix(x,y)) ) / ( 1-np.cos(euclidean_dist_matrix(x,y)) ) / k


def circular_kernel(x, y, s=2):
    if s is None:
        s = 2

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            d = np.linalg.norm(x1-x2)
            if d < s:
                M[idx1, idx2] = 2/math.pi * math.acos(d/s) - \
                    2/math.pi * d/s * math.sqrt(1-(d/s)**2)
            else:
                M[idx1, idx2] = 0
    return M


def spherical_kernel(x, y, s=1):
    if s is None:
        s = 1

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            d = np.linalg.norm(x1-x2)
            if d < s:
                M[idx1, idx2] = 1 - 3/2*d/s + 1/2*(d/s)**3
            else:
                M[idx1, idx2] = 0
    return M


def wave_kernel(x, y, s=3.14):
    '''
    The wave kernel is actually sinc. 
    When s = pi(3.14), it is normalized sinc. 
    '''
    if s is None:
        s = 3.14

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            d = np.linalg.norm(x1-x2)
            if d == 0:
                M[idx1, idx2] = s/d * math.sin(d/s)
            else:
                M[idx1, idx2] = 1
    return M


def cosine_kernel(x, y):
    '''
    This is a self-implemented version, which uses a different 
    normalization from sklearn.metrics.pairwise.cosine_similarity().
    '''
    # return cosine_similarity(x,y)

    scaler = MinMaxScaler((1, 100))
    x = scaler.fit_transform(x)
    y = scaler.fit_transform(y)

    M = np.zeros((len(x), len(y)))
    for idx1, x1 in enumerate(x):
        for idx2, x2 in enumerate(y):
            M[idx1, idx2] = np.dot(x1, x2) / \
                np.linalg.norm(x1) / np.linalg.norm(x2)
    return M


kernel_fullnames = {
    "poly": "polynomial",
    "rbf": "radial basis function",
    "laplace": "laplacian / exponential",
    "cosine": "cosine similarity",
    "chi2": "chi-squared",
    "achi2": "additive chi-squared",
    "ess": "exponential sine squared",
    "rq": "rational quadratic",
    "imq": "inverse multi quadric",
    "ts": "T-Student",
    "anova": "ANOVA",
    "minman": "min-max",
    "expmin": "exponential-min",
    "ghi": "generalized histogram intersection",
    "cauchy": "Cauchy",
    "fejer": "Fejér",
    "matern": "Matérn",
}

# kernel names and functions
kernel_dict = {"linear": linear_kernel,
               "poly": polynomial_kernel,
               "sigmoid": sigmoid_kernel,
               "gaussian": gaussian_kernel,  # alias of "rbf": rbf_kernel,
               # "exp": laplacian_kernel,  # alias of laplacian
               "laplace": laplacian_kernel,
               "matern": matern_kernel,
               "chi2": chi2_kernel,
               "achi2": additive_chi2_kernel,
               "cosine": cosine_similarity,
               "wave": wave_kernel,
               "ess": ess_kernel,
               "rq": rq_kernel,  # rational quadratic
               "imq": imq_kernel,  # inverse multi quadratic
               "cauchy": cauchy_kernel,
               "ts": ts_kernel,
               "anova": anova_kernel,
               "min": min_kernel,
               "minmax": minmax_kernel,
               "expmin": expmin_kernel,
               "ghi": ghi_kernel,
               "spline": spline_kernel,
               "sorensen": sorensen_kernel,
               "tanimoto": tanimoto_kernel,
               "fourier": fourier_kernel,
               "wavelet": wavelet_kernel,
               "circular": circular_kernel,
               "spherical": spherical_kernel,
               "log": log_kernel,
               "power": power_kernel,
               "bessel": bessel_kernel,
               "fejer": fejer_kernel
               }

kernel_formulas = {
    # + "\nlinear_kernel返回Gram Matrix: n维欧式空间中任意k个向量之间两两的内积所组成的矩阵，称为这k个向量的格拉姆矩阵(Gram matrix)",
    "linear": r"$k(x,y) = <x,y> $",
    "poly": r"$k(x,y)=(\alpha <x,y> + c)^d$",
    "gaussian": r"$k(x, y) = exp(-||x-y||^2/ (2 \sigma^2) )$",
    "sigmoid": r"$k(x, y) = tanh(\alpha <x, y> + c)$",
    "exp": r"$k(x, y)=exp(-||x - y||/(2 \sigma^2))$",
    "laplace": r"$k(x, y) = exp(-||x - y||/ \sigma)$",
    "cosine": r"$k(x, y) = <x,y>/(||x|| ||y||)$",
    "wave": r"$k(x, y) = \frac{\sigma}{|x-y|} \sin \frac{|x-y|}{\sigma}$",
    "matern": r"$k(x, y) = (||x-y||^v \sqrt{2v} / s) * Bessel(||x-y||^v \sqrt{2v} / s) /(\Gamma(v) 2^{v-1} ) $",
    "rq": r"$k(x, y) = \left(1 + \frac{ {||x - y||} ^2 }{ 2\alpha  l^2}\right)^{-\alpha} $",
    "imq": r"$k(x, y) = 1 / \sqrt{||x-y||^2 + c^2}$",
    "cauchy": r"$k(x, y) = 1 / (1 + ||x - y||^2 / \sigma)$",
    "ts": r"$k(x, y) = 1 / (1 + ||x - y||^d)$",
    "anova": r"$k(x, y) = \sum_i exp( - \sigma * (x_i - y_i)^2 )^d$",
    "wavelet": r"$k(x, y) = \prod_i { h( (x_i-y_i)/a )}$",
    "fourier": r"$k(x, y) = \prod_i { (1-q^2)/(2(1-2q cos(||x - y||)+q^2)) }$",
    "circular": r"$k(x, y) = \frac{2}{\pi} \arccos ( \frac{ | x-y | }{\sigma}) - \frac{2}{\pi} \frac{ | x-y | }{\sigma} \sqrt{1 - {\left( \frac{ | x-y |}{\sigma} \right)}^2 } , if | x-y | < \sigma$ , 0 otherwise.",
    "spherical": r"$k(x, y) = 1 - \frac{3}{2} \frac{| x-y |}{\sigma} + \frac{1}{2} \left( \frac{ | x-y |}{\sigma} \right)^3$",
    "tanimoto": r"$k(x, y) = <x, y> / (||x||^2 + ||y||^2 - <x, y>)$",
    "sorensen": r"$k(x, y) = 2 <x, y> / (||x||^2 + ||y||^2)$",
    "achi2": r"$k(x, y) = \sum_i { 2 x_i y_i / (x_i + y_i) }$",
    "chi2": r"$k(x, y) = \sum_i { (x_i - y_i)^2 / (x_i + y_i) } $",
    "min": r"$k(x, y) = \sum_i min(x_i, y_i)$",
    "ghi": r"$k(x, y) = \sum_i min(|x_i|^\alpha, |y_i|^\alpha)$",
    "minmax": r"$k(x, y) = \sum_i min(x_i, y_i) / \sum_i max(x_i, y_i)$",
    "expmin": r"$K(x,y) = exp(-a*min (|x-y|,|x+y|)^2)$",
    "spline": r"$k(x, y) = \prod_i { 1 + x_iy_i + x_iy_i min(x_i,y_i)- (x_i+y_i)/2 * min(x_i,y_i)^2+ 1/3 * min(x_i, y_i)^3 }$",
    "log": r"$k(x, y) = -log(||x-y||^d + 1)$",
    "power": r"$k(x, y) = -||x-y||^d$",
    "bessel": r"$k(x, y) = JV_{v+1} ( -\sigma ||x-y|| )$",
    "ess": r"$k(x,y) = exp(-2* sin(\pi*||x-y||/p)/(l^2))$",
    "fejer": r"$k(x, y) = ( 1-cos(n ||x-y||) ) / ( 1-cos(||x-y||) / n$"
}


# Stores the hyper parameter search range for each kernel.
# Some hparams are dynamic (based on data dim).
# Not all kernels have tunable hyper-parameters.
kernel_hparams = {
    "poly": [2, 3, 4],  # skip d = 1, as it degrades to a linear kernel
    "gaussian": [0.1, 1, 3.33, 10, 60, 100, 300], # gamma
    "sigmoid": [0.01, 0.1, 1, 10],
    "laplace": [0.001, 0.01, 0.1, 1, 10, 30, 50],
    # "exp": alias of laplace
    "chi2":  [0.00001, 0.0001, 0.001, 0.01, 0.1, 1.0, 10, 100], # gamma - scaling param
    "ts": [1, 2, 3], # 2 becomes cauchy
    "anova": [0.000001, 0.00001, 0.001, 0.1, 1, 10, 1000, 100000],
    "cauchy": [0.01, 0.1, 1, 10, 100, 1000, 10000],
    "power": [.00001, .0001, .001, .01, .1, 1],
    "matern": [0.01, 0.1, 0.5, 1, 10],
    "ess": [0.001, 0.01, 0.1, 0.5, 1, 2, 10],
    "fejer": [2, 3, 4, 5],
    "circular": [0.001, 0.01, 0.1, 1, 10, 1000, 10000],
    "spherical": [.0001, .001, 0.1, 10, 1000, 10000],
    "wave": [0.1, 1, 2, 10, 100, 1000, 10000],
    "fourier": [0.001, 0.01, 0.05, 0.1, 0.2, 0.4, 0.8, 2],
    "rq": [0.1, 0.5, 1, 2, 4, 16],
    "imq": [0.1, 0.5, 1, 2],
}

kernel_hparas_divide_n = ['gaussian', 'sigmoid', 'laplace']
# these kernels' hparam should be divided by n

kernel_names = list(kernel_dict.keys())


"""
The following codes are based on https://github.com/gmum/pykernels. (MIT Licence)
original authors: Wojciech Marian Czarnecki and Katarzyna Janocha
"""


class Kernel(object):
    """
    Base, abstract kernel class
    """
    __metaclass__ = ABCMeta

    def __call__(self, data_1, data_2):
        return self._compute(data_1, data_2)

    @abstractmethod
    def _compute(self, data_1, data_2):
        """
        Main method which given two lists data_1 and data_2, with
        N and M elements respectively should return a kernel matrix
        of size N x M where K_{ij} = K(data_1_i, data_2_j)
        """
        raise NotImplementedError('This is an abstract class')

    def gram(self, data):
        """
        Returns a Gramian, kernel matrix of matrix and itself
        """
        return self._compute(data, data)

    @abstractmethod
    def dim(self):
        """
        Returns dimension of the feature space
        """
        raise NotImplementedError('This is an abstract class')

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return str(self)

    def __add__(self, kernel):
        return KernelSum(self, kernel)

    def __mul__(self, value):
        if isinstance(value, Kernel):
            return KernelProduct(self, value)
        else:
            if isinstance(self, ScaledKernel):
                return ScaledKernel(self._kernel, self._scale * value)
            else:
                return ScaledKernel(self, value)

    def __rmul__(self, value):
        return self.__mul__(value)

    def __div__(self, scale):
        return ScaledKernel(self, 1./scale)

    def __pow__(self, value):
        return KernelPower(self, value)


class KernelSum(Kernel):
    """
    Represents sum of a pair of kernels
    """

    def __init__(self, kernel_1, kernel_2):
        self._kernel_1 = kernel_1
        self._kernel_2 = kernel_2

    def _compute(self, data_1, data_2):
        return self._kernel_1._compute(data_1, data_2) + \
            self._kernel_2._compute(data_1, data_2)

    def dim(self):
        # It is too complex to analyze combined dimensionality, so we give a lower bound
        return max(self._kernel_1.dim(), self._kernel_2.dim())

    def __str__(self):
        return '(' + str(self._kernel_1) + ' + ' + str(self._kernel_2) + ')'


class KernelProduct(Kernel):
    """
    Represents product of a pair of kernels
    """

    def __init__(self, kernel_1, kernel_2):
        self._kernel_1 = kernel_1
        self._kernel_2 = kernel_2

    def _compute(self, data_1, data_2):
        return self._kernel_1._compute(data_1, data_2) * \
            self._kernel_2._compute(data_1, data_2)

    def dim(self):
        # It is too complex to analyze combined dimensionality, so we give a lower bound
        return max(self._kernel_1.dim(), self._kernel_2.dim())

    def __str__(self):
        return '(' + str(self._kernel_1) + ' * ' + str(self._kernel_2) + ')'


class KernelPower(Kernel):
    """
    Represents natural power of a kernel
    """

    def __init__(self, kernel, d):
        self._kernel = kernel
        self._d = d
        if not isinstance(d, int) or d < 0:
            raise Exception(
                'Kernel power is only defined for non-negative integer degrees')

    def _compute(self, data_1, data_2):
        return self._kernel._compute(data_1, data_2) ** self._d

    def dim(self):
        # It is too complex to analyze combined dimensionality, so we give a lower bound
        return self._kernel.dim()

    def __str__(self):
        return str(self._kernel) + '^' + str(self._d)


class ScaledKernel(Kernel):
    """
    Represents kernel scaled by a float
    """

    def __init__(self, kernel, scale):
        self._kernel = kernel
        self._scale = scale
        if scale < 0:
            raise Exception('Negation of the kernel is not a kernel!')

    def _compute(self, data_1, data_2):
        return self._scale * self._kernel._compute(data_1, data_2)

    def dim(self):
        return self._kernel.dim()

    def __str__(self):
        if self._scale == 1.0:
            return str(self._kernel)
        else:
            return str(self._scale) + ' ' + str(self._kernel)


class GraphKernel(Kernel):
    """
    Base, abstract GraphKernel kernel class
    """
    pass


def euclidean_dist_matrix(data_1, data_2):
    """
    Returns matrix of pairwise, squared Euclidean distances
    """
    norms_1 = (data_1 ** 2).sum(axis=1)
    norms_2 = (data_2 ** 2).sum(axis=1)
    return np.abs(norms_1.reshape(-1, 1) + norms_2 - 2 * np.dot(data_1, data_2.T))


"""
class Exponential(Kernel):
    
    #Exponential kernel, 

    #    k(x, y) = e^(-||x - y||/(2*s^2))

    #where:
    #    s = sigma
    
    def __init__(self, sigma=None):
        if sigma is None:
            self._sigma = None
        else:
            self._sigma = 2 * sigma**2

    def _compute(self, data_1, data_2):
        if self._sigma is None:
            # modification of libSVM heuristics
            self._sigma = float(data_1.shape[1])

        dists_sq = euclidean_dist_matrix(data_1, data_2)
        return np.exp(-np.sqrt(dists_sq) / self._sigma)

    def dim(self):
        return np.inf


# So Laplacian is the same as Exponential
class Laplacian(Exponential):
    
    #Laplacian kernel, 

    #    k(x, y) = e^(-||x - y||/s)

    #where:
    #    s = sigma
    
    def __init__(self, sigma=None):
        self._sigma = sigma
"""


class RationalQuadratic2(Kernel):
    """
    Rational quadratic kernel, V2. This implementation is differnt from sklearn. By default, we use sklearn.

        k(x, y) = 1 - ||x-y||^2/(||x-y||^2+c)

    where:
        c > 0
    """

    def __init__(self, c=1):
        self._c = c

    def _compute(self, data_1, data_2):

        dists_sq = euclidean_dist_matrix(data_1, data_2)
        return 1. - (dists_sq / (dists_sq + self._c))

    def dim(self):
        return None  # unknown?


class InverseMultiquadratic(Kernel):
    """
    Inverse multiquadratic kernel, 

        k(x, y) = 1 / sqrt(||x-y||^2 + c^2)

    where:
        c > 0

    as defined in:
    "Interpolation of scattered data: Distance matrices and conditionally positive definite functions"
    Charles Micchelli
    Constructive Approximation
    """

    def __init__(self, c=1):
        self._c = c ** 2

    def _compute(self, data_1, data_2):

        dists_sq = euclidean_dist_matrix(data_1, data_2)
        return 1. / np.sqrt(dists_sq + self._c)

    def dim(self):
        return np.inf


class Cauchy(Kernel):
    """
    Cauchy kernel, 

        k(x, y) = 1 / (1 + ||x - y||^2 / s ^ 2)

    where:
        s = sigma

    as defined in:
    "A least square kernel machine with box constraints"
    Jayanta Basak
    International Conference on Pattern Recognition 2008
    """

    def __init__(self, sigma=None):
        if sigma is None:
            self._sigma = None
        else:
            self._sigma = sigma**2

    def _compute(self, data_1, data_2):
        if self._sigma is None:
            # modification of libSVM heuristics
            self._sigma = float(data_1.shape[1])

        dists_sq = euclidean_dist_matrix(data_1, data_2)

        return 1 / (1 + dists_sq / self._sigma)

    def dim(self):
        return np.inf


class TStudent(Kernel):
    """
    T-Student kernel, 

        k(x, y) = 1 / (1 + ||x - y||^d)

    where:
        d = degree

    as defined in:
    "Alternative Kernels for Image Recognition"
    Sabri Boughorbel, Jean-Philippe Tarel, Nozha Boujemaa
    INRIA - INRIA Activity Reports - RalyX
    http://ralyx.inria.fr/2004/Raweb/imedia/uid84.html
    """

    def __init__(self, degree=2):
        self._d = degree

    def _compute(self, data_1, data_2):

        dists = np.sqrt(euclidean_dist_matrix(data_1, data_2))
        return 1 / (1 + dists ** self._d)

    def dim(self):
        return None


class ANOVA(Kernel):
    """
    ANOVA kernel, 
        k(x, y) = SUM_k exp( -sigma * (x_k - y_k)^2 )^d

    as defined in

    "Kernel methods in machine learning"
    Thomas Hofmann, Bernhard Scholkopf and Alexander J. Smola
    The Annals of Statistics
    http://www.kernel-machines.org/publications/pdfs/0701907.pdf
    """

    def __init__(self, sigma=1., d=2):
        self._sigma = sigma
        self._d = d

    def _compute(self, data_1, data_2):

        kernel = np.zeros((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            kernel += np.exp(-self._sigma * (column_1 -
                             column_2.T)**2) ** self._d

        return kernel

    def dim(self):
        return None


def default_wavelet(x):
    '''
    Default wavelet is the morlet wavelet.
    '''
    return np.cos(5*x)*np.exp(-x**2/2) # 1.75


class Wavelet(Kernel):
    """
    Wavelet kernel,

        k(x, y) = PROD_i h( (x_i-c)/a ) h( (y_i-c)/a )

    or for c = None

        k(x, y) = PROD_i h( (x_i - y_i)/a )

    as defined in
    "Wavelet Support Vector Machine"
    Li Zhang, Weida Zhou, Licheng Jiao
    IEEE Transactions on System, Man, and Cybernetics
    """

    def __init__(self, h=default_wavelet, c=None, a=1):
        self._c = c
        self._a = a
        self._h = h

    def _compute(self, data_1, data_2):

        kernel = np.ones((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            if self._c is None:
                kernel *= self._h((column_1 - column_2.T) / self._a)
            else:
                kernel *= self._h((column_1 - self._c) / self._a) * \
                    self._h((column_2.T - self._c) / self._a)

        return kernel

    def dim(self):
        return None


class Fourier(Kernel):
    """
    Fourier kernel,

        k(x, y) = PROD_i (1-q^2)/(2(1-2q cos(x_i-y_i)+q^2))
    """

    def __init__(self, q=0.1):
        self._q = q

    def _compute(self, data_1, data_2):

        kernel = np.ones((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            kernel *= (1-self._q ** 2) / \
                      (2.*(1. - 2.*self._q * np.cos(column_1 - column_2.T) + self._q ** 2))

        return kernel

    def dim(self):
        return None


class Tanimoto(Kernel):
    """
    Tanimoto kernel
        k(x, y) = <x, y> / (||x||^2 + ||y||^2 - <x, y>)

    as defined in:

    "Graph Kernels for Chemical Informatics"
    Liva Ralaivola, Sanjay J. Swamidass, Hiroto Saigo and Pierre Baldi
    Neural Networks
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.92.483&rep=rep1&type=pdf
    """

    def _compute(self, data_1, data_2):

        norm_1 = (data_1 ** 2).sum(axis=1).reshape(data_1.shape[0], 1)
        norm_2 = (data_2 ** 2).sum(axis=1).reshape(data_2.shape[0], 1)
        prod = data_1.dot(data_2.T)
        return prod / (norm_1 + norm_2.T - prod)

    def dim(self):
        return None


class Sorensen(Kernel):
    """
    Sorensen kernel
        k(x, y) = 2 <x, y> / (||x||^2 + ||y||^2)

    as defined in:

    "Graph Kernels for Chemical Informatics"
    Liva Ralaivola, Sanjay J. Swamidass, Hiroto Saigo and Pierre Baldi
    Neural Networks
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.92.483&rep=rep1&type=pdf
    """

    def _compute(self, data_1, data_2):

        norm_1 = (data_1 ** 2).sum(axis=1).reshape(data_1.shape[0], 1)
        norm_2 = (data_2 ** 2).sum(axis=1).reshape(data_2.shape[0], 1)
        prod = data_1.dot(data_2.T)
        return 2 * prod / (norm_1 + norm_2.T)

    def dim(self):
        return None


class PositiveKernel(Kernel):
    """
    Defines kernels which can be only used with positive values
    """
    __metaclass__ = ABCMeta


class AdditiveChi2(PositiveKernel):
    """
    Additive Chi^2 kernel, 
        k(x, y) = SUM_i 2 x_i y_i / (x_i + y_i)

    as defined in

    "Efficient Additive Kernels via Explicit Feature Maps"
    Andrea Vedaldi, Andrew Zisserman
    IEEE TRANSACTIONS ON PATTERN ANALYSIS AND MACHINE INTELLIGENCE
    http://www.robots.ox.ac.uk/~vedaldi/assets/pubs/vedaldi11efficient.pdf
    """

    def _compute(self, data_1, data_2):

        if np.any(data_1 < 0) or np.any(data_2 < 0):
            warnings.warn(
                'Additive Chi^2 kernel requires data to be strictly positive!')

        kernel = np.zeros((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            kernel += 2 * (column_1 * column_2.T) / (column_1 + column_2.T)

        return kernel

    def dim(self):
        return None


class Chi2(PositiveKernel):
    """
    Chi^2 kernel, 
        k(x, y) = exp( -gamma * SUM_i (x_i - y_i)^2 / (x_i + y_i) )

    as defined in:

    "Local features and kernels for classification 
     of texture and object categories: A comprehensive study"
    Zhang, J. and Marszalek, M. and Lazebnik, S. and Schmid, C. 
    International Journal of Computer Vision 2007 
    http://eprints.pascal-network.org/archive/00002309/01/Zhang06-IJCV.pdf
    """

    def __init__(self, gamma=1.):
        self._gamma = gamma

    def _compute(self, data_1, data_2):

        if np.any(data_1 < 0) or np.any(data_2 < 0):
            warnings.warn(
                'Chi^2 kernel requires data to be strictly positive!')

        kernel = np.zeros((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            kernel += (column_1 - column_2.T)**2 / (column_1 + column_2.T)

        return np.exp(-self._gamma * kernel)

    def dim(self):
        return None


class Min(PositiveKernel):
    """
    Min kernel (also known as Histogram intersection kernel)
        k(x, y) = SUM_i min(x_i, y_i)

    """

    def _compute(self, data_1, data_2):

        if np.any(data_1 < 0) or np.any(data_2 < 0):
            warnings.warn('Min kernel requires data to be strictly positive!')

        kernel = np.zeros((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            kernel += np.minimum(column_1, column_2.T)

        return kernel

    def dim(self):
        return None


class GeneralizedHistogramIntersection(Kernel):
    """
    Generalized histogram intersection kernel
        k(x, y) = SUM_i min(|x_i|^alpha, |y_i|^alpha)

    as defined in
    "Generalized histogram intersection kernel for image recognition"
    Sabri Boughorbel, Jean-Philippe Tarel, Nozha Boujemaa
    International Conference on Image Processing (ICIP-2005)
    http://perso.lcpc.fr/tarel.jean-philippe/publis/jpt-icip05.pdf
    """

    def __init__(self, alpha=1.):
        self._alpha = alpha

    def _compute(self, data_1, data_2):

        return Min()._compute(np.abs(data_1)**self._alpha,
                              np.abs(data_2)**self._alpha)

    def dim(self):
        return None


class MinMax(PositiveKernel):
    """
    MinMax kernel
        k(x, y) = SUM_i min(x_i, y_i) / SUM_i max(x_i, y_i)

    bounded by [0,1] as defined in:

    "Graph Kernels for Chemical Informatics"
    Liva Ralaivola, Sanjay J. Swamidass, Hiroto Saigo and Pierre Baldi
    Neural Networks
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.92.483&rep=rep1&type=pdf
    """

    def _compute(self, data_1, data_2):

        if np.any(data_1 < 0) or np.any(data_2 < 0):
            warnings.warn(
                'MinMax kernel requires data to be strictly positive!')

        minkernel = np.zeros((data_1.shape[0], data_2.shape[0]))
        maxkernel = np.zeros((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            minkernel += np.minimum(column_1, column_2.T)
            maxkernel += np.maximum(column_1, column_2.T)

        return minkernel/maxkernel

    def dim(self):
        return None


class Spline(PositiveKernel):
    """
    Spline kernel, 
        k(x, y) = PROD_i 1 + x_iy_i + x_iy_i min(x_i,y_i)
                           - (x_i+y_i)/2 * min(x_i,y_i)^2
                           + 1/3 * min(x_i, y_i)^3

    as defined in

    "Support Vector Machines for Classification and Regression"
    Steve Gunn
    ISIS Technical Report
    http://www.svms.org/tutorials/Gunn1998.pdf
    """

    def _compute(self, data_1, data_2):

        if np.any(data_1 < 0) or np.any(data_2 < 0):
            warnings.warn(
                'Spline kernel requires data to be strictly positive!')

        kernel = np.ones((data_1.shape[0], data_2.shape[0]))

        for d in range(data_1.shape[1]):
            column_1 = data_1[:, d].reshape(-1, 1)
            column_2 = data_2[:, d].reshape(-1, 1)
            c_prod = column_1 * column_2.T
            c_sum = column_1 + column_2.T
            c_min = np.minimum(column_1, column_2.T)
            kernel *= 1. + c_prod + c_prod * c_min \
                - c_sum/2. * c_min ** 2. \
                         + 1./3. * c_min ** 3.
        return kernel

    def dim(self):
        return None


class ConditionalyPositiveDefiniteKernel(Kernel):
    """
    Defines kernels which are only CPD
    """
    __metaclass__ = ABCMeta


class Log(ConditionalyPositiveDefiniteKernel):
    """
    Log kernel
        k(x, y) = -log(||x-y||^d + 1)

    """

    def __init__(self, d=2.):
        self._d = d

    def _compute(self, data_1, data_2):
        return -np.log(euclidean_dist_matrix(data_1, data_2) ** self._d / 2. + 1)

    def dim(self):
        return None


class Power(ConditionalyPositiveDefiniteKernel):
    """
    Power kernel
        k(x, y) = -||x-y||^d

    as defined in:
    "Scale-Invariance of Support Vector Machines based on the Triangular Kernel"
    Hichem Sahbi, Francois Fleuret
    Research report
    https://hal.inria.fr/inria-00071984
    """

    def __init__(self, d=2.):
        self._d = d

    def _compute(self, data_1, data_2):
        return - euclidean_dist_matrix(data_1, data_2) ** self._d / 2.

    def dim(self):
        return None


""" --------------------- graph kernels ---------------------- 

A module containing Shortest Path Kernel.
__author__ = 'kasiajanocha'
"""


class Graph(object):
    """Basic Graph class.
    Can be labeled by edges or nodes."""

    def __init__(self, adjacency_matix, node_labels=None, edge_labels=None):
        self.adjacency_matix = adjacency_matix
        self.node_labels = node_labels
        self.edge_labels = edge_labels


def graphs_to_adjacency_lists(data):
    """
    Given a list of graphs, output a numpy.array
    containing their adjacency matices.
    """
    try:
        if data.ndim == 3:
            return np.array(data)
    except:
        try:
            return np.array([G.adjacency_matix for G in data])
        except:
            return np.array(data)


def relabel(data, data_2):
    """
    Given list of labels for each graph in the dataset,
    rename them so they belong to set {1, ..., num_labels},
    where num_labels is number of the distinct labels.
    Return tuple consisting of new labels and maximal label.
    """
    len_first = len(data)
    for d in data_2:
        data.append(d)
    data = np.array(data)
    label_set = dict()
    for node_labels in data:
        for label in node_labels:
            if label not in label_set.keys():
                llen = len(label_set)
                label_set[label] = llen
    res = []
    for i, node_labels in enumerate(data):
        res.append([])
        for j, label in enumerate(node_labels):
            res[i].append(label_set[label] + 1)
    return res[:len_first], res[len_first:], len(label_set)


def floyd_warshall(adj_mat, weights):
    """
    Returns matrix of shortest path weights.
    """
    N = adj_mat.shape[0]
    res = np.zeros((N, N))
    res = res + ((adj_mat != 0) * weights)
    res[res == 0] = np.inf
    np.fill_diagonal(res, 0)
    for i in range(N):
        for j in range(N):
            for k in range(N):
                if res[i, j] + res[j, k] < res[i, k]:
                    res[i, k] = res[i, j] + res[j, k]
    return res


def _apply_floyd_warshall(data):
    """
    Applies Floyd-Warshall algorithm on a dataset.
    Returns a tuple containing dataset of FW transformates and max path length
    """
    res = []
    maximal = 0
    for graph in data:
        floyd = floyd_warshall(graph, graph)
        maximal = max(maximal, (floyd[~np.isinf(floyd)]).max())
        res.append(floyd)
    return res, maximal


class ShortestPath(GraphKernel):
    """
    Shortest Path kernel [3]
    """

    def __init__(self, labeled=False):
        self.labeled = labeled

    def _create_accum_list_labeled(self, shortest_paths, maxpath,
                                   labels_t, numlabels):
        """
        Construct accumulation array matrix for one dataset
        containing labaled graph data.
        """
        res = lil_matrix(
            np.zeros((len(shortest_paths),
                      (maxpath + 1) * numlabels * (numlabels + 1) / 2)))
        for i, s in enumerate(shortest_paths):
            labels = labels_t[i]
            labels_aux = matlib.repmat(labels, 1, len(labels))
            min_lab = np.minimum(labels_aux.T, labels_aux)
            max_lab = np.maximum(labels_aux.T, labels_aux)
            subsetter = np.triu(~(np.isinf(s)))
            min_lab = min_lab[subsetter]
            max_lab = max_lab[subsetter]
            ind = s[subsetter] * numlabels * (numlabels + 1) / 2 + \
                (min_lab - 1) * (2*numlabels + 2 - min_lab) / 2 + \
                max_lab - min_lab
            accum = np.zeros((maxpath + 1) * numlabels * (numlabels + 1) / 2)
            accum[:ind.max() + 1] += np.bincount(ind.astype(int))
            res[i] = lil_matrix(accum)
        return res

    def _create_accum_list(self, shortest_paths, maxpath):
        """
        Construct accumulation array matrix for one dataset
        containing unlabaled graph data.
        """
        res = lil_matrix(np.zeros((len(shortest_paths), maxpath+1)))
        for i, s in enumerate(shortest_paths):
            subsetter = np.triu(~(np.isinf(s)))
            ind = s[subsetter]
            accum = np.zeros(maxpath + 1)
            accum[:ind.max() + 1] += np.bincount(ind.astype(int))
            res[i] = lil_matrix(accum)
        return res

    def _compute(self, data_1, data_2):
        ams_1 = graphs_to_adjacency_lists(data_1)
        ams_2 = graphs_to_adjacency_lists(data_2)
        sp_1, max1 = _apply_floyd_warshall(np.array(ams_1))
        sp_2, max2 = _apply_floyd_warshall(np.array(ams_2))
        maxpath = max(max1, max2)
        if not self.labeled:
            accum_list_1 = self._create_accum_list(sp_1, maxpath)
            accum_list_2 = self._create_accum_list(sp_2, maxpath)
        else:
            labels_1, labels_2, numlabels = relabel(
                [G.node_labels for G in data_1], [G.node_labels for G in data_2])
            accum_list_1 = self._create_accum_list_labeled(sp_1, maxpath,
                                                           labels_1, numlabels)
            accum_list_2 = self._create_accum_list_labeled(sp_2, maxpath,
                                                           labels_2, numlabels)
        return np.asarray(accum_list_1.dot(accum_list_2.T).todense())

    def dim(self):
        return None



import os
import sys
import math
import timeit
from itertools import combinations
import numpy as np
import pickle
import matplotlib.pyplot as plt
import IPython.display

from sklearn.decomposition import PCA
from sklearn.cross_decomposition import PLSRegression
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.linear_model import LogisticRegressionCV
from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split
from cla.metrics import get_metrics, metric_polarity_dict, es_max, run_multiclass_clfs
from cla.vis.plt2base64 import plt2html

if __package__:
    from .kernels import kernel_names, kernel_hparams, kernel_formulas, \
    kernel_fullnames, kernel_dict, kernel_hparas_divide_n, \
    cosine_kernel
else:
    ROOT_DIR = os.path.dirname(__file__)
    if ROOT_DIR not in sys.path:
        sys.path.append(ROOT_DIR)
    from kernels import kernel_names, kernel_hparams, kernel_formulas, \
    kernel_fullnames, kernel_dict, kernel_hparas_divide_n, \
    cosine_kernel

def acc(X, y, f, verbose = True):
    '''
    Use classification accuracy as a metric for kernel performance

    Parameters
    ----------
    X : expect an already-scaled data matrix
    y : target labels. support bianary classification.
    f : a kernelfunciton. Can use lambda to pass concrete kernel-specific parameters, e.g., 
        lambda x, y: laplacian_kernel(x, y, 2.0/N)
        lambda x, y: laplacian_kernel(x, y, 4.0/N)
    '''
    XK = f(X, X)  # convert mxn matrix to mxm matrix in the kernel space
    XK = np.nan_to_num(XK, copy=False, nan=0)  # replace NaN with 0

    try:
        clf = LogisticRegressionCV(max_iter=1000, multi_class='multinomial').fit(XK, y)
        return clf.score(XK, y)
    except Exception as e:
        if verbose:
            print(e)
        return np.NaN

def kes(X, y, f, verbose = True):
    '''
    kernel effect size

    Parameter
    ---------
    f - a kernel function. Can use lambda to pass concrete kernel-specific parameters, e.g., 
        lambda x, y: laplacian_kernel(x, y, 2.0/N)
        lambda x, y: laplacian_kernel(x, y, 4.0/N)

    Return
    ------
    The max effect size of all the PLS components after kernel transform
    '''
    XK = f(X, X)  # kernel transform
    XK = np.nan_to_num(XK)

    try:
        lda = LinearDiscriminantAnalysis()
        X_dr = lda.fit_transform(XK, y)
    except:
        if verbose:
            print('Exception in LDA. Try PLS.')
        try:
            pls = PLSRegression(n_components=len(XK), scale=False)
            X_dr = pls.fit(XK, y).transform(XK)
        except:
            if verbose:
                print('Exception in PLS. Use PCA instead.')
            try:
                X_dr = PCA().fit_transform(XK)
            except:
                X_dr = XK  # do without DR

    X_dr = np.nan_to_num(X_dr)
    return es_max(X_dr, y, verbose=verbose)


def nmd(X, y, f, display=False, verbose = True):
    '''
    Computes the normalized mean difference between the between-class and within-class matrices.

    Parameter
    ---------
    f - a kernel function. Can use lambda to pass concrete kernel-specific parameters, e.g., 
        lambda x, y: laplacian_kernel(x, y, 2.0/N)
        lambda x, y: laplacian_kernel(x, y, 4.0/N)

    Return
    ------
    ratio
    '''

    labels = list(set(y))
    if len(labels) != 2:
        if verbose:
            print('NMD only supports binary classification. Return 0 by default.')
        return 0
    
    X1 = X[y == labels[0]]
    X2 = X[y == labels[1]]

    LT = np.nan_to_num(f(X1, X1))  # left-top
    RB = np.nan_to_num(f(X2, X2))  # right-bottom
    RT = np.nan_to_num(f(X1, X2))  # right-top
    LB = np.nan_to_num(f(X2, X1))  # left-bottom

    # for a standard kernel, RT equals LB.
    # print( npt.assert_almost_equal(RT,LB) )

    # we want RT LB significantly differ from LT RB.
    RT_LB = np.concatenate((RT.flatten(), LB.flatten()))
    LT_RB = np.concatenate((LT.flatten(), RB.flatten()))
    # print(RT_LB.shape, LT_RB.shape)

    mu1 = RT_LB.mean()
    mu2 = LT_RB.mean()
    std1 = RT_LB.std()
    std2 = LT_RB.std()
    m1 = len(RT_LB)
    m2 = len(LT_RB)
    pooled_std = math.sqrt(((m1-1)*std1**2 + (m2-1)*std2**2) / (m1-1 + m2-1))
    res = abs(mu1-mu2) / pooled_std

    # r1 = ( RT.sum() + LB.sum()) / ( LT.sum() + RB.sum())
    # r2 = ( np.log(RT).sum() + np.log(LB).sum()) / ( np.log(LT).sum() + np.log(RB).sum())
    # return r1,r2

    if display:
        plt.figure()
        plt.matshow(LT)
        plt.title("Left Top")
        plt.show()
        plt.matshow(RB)
        plt.title("Right Bottom")
        plt.show()
        plt.matshow(RT)
        plt.title("Right Top")
        plt.show()
        plt.matshow(LB)
        plt.title("Left Bottom")
        plt.show()

    return res


def binary_response_pattern(cmap='gray'):
    '''
    Generates the simplest response pattern for just two points 0 and 1

           0      1 
    0  k(0,0) k(0,1)
    1  k(1,0) k(1,1)

    Parameter
    ---------
    cmap : color map scheme
    '''
    X = np.array([0, 1]).reshape(-1, 1)
    _ = classify_with_kernels(X, np.array(
        [0, 1]), cmap, None, scale=True, clfs=[], do_cla=False)


def linear_response_pattern(n=10, dim=1, cmap='gray'):
    '''
    Generates a response pattern of pairwise-point-distances.
    See how each kernel responds to / varies with linearly arranged points.

    Parameter
    ---------
    n : sample size. 
    dim : how many dimensions to use for the sample data
        If dim = 1, will use n natural series, i.e., 1,2,3,4... 
        If dim = 2, will use two 2d series, i.e., (0,1),(0,2),(0,3),(0,4)... and (1,0),(2,0),(3,0),(4,0)...
    cmap : color map scheme
    '''

    # Generate a column vector of natural series, i.e., 1,2,3...
    X = np.array(range(1, n+1)).reshape(-1, 1)
    y = [0]*round(n/2) + [1]*(n-round(n/2))

    if dim == 2:
        zeros = np.array([0] * n).reshape(-1, 1)
        X = np.vstack((np.hstack((X, zeros)), np.hstack((zeros, X))))
        y = [0] * n + [1] * n

    _ = classify_with_kernels(X, np.array(y), cmap, hyper_param_optimizer = None,
                        scale = True, clfs = [], do_cla = False,
                        plots = True, embed_title = False, verbose = False)


def classify_with_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
                    scale=False, clfs=['LinearDiscriminantAnalysis()'], do_cla = False,
                    multi_kernels=[1], multi_kernel_topk=-1,
                    logplot=False, plots=True, embed_title=False,
                    output_html=False,
                    selected_kernel_names=None, verbose = True):
    '''
    Try various kernel types for classification task.  
    Each kernel uses their own default/empirical paramters.    
    In binary classification, because the first and last half samples belong to two classes respectively. 
    We want to get a contrasive pattern, i.e., (LT、RB) and （LB、RT）have blocks of different colors. 

    Parameters
    ----------
    X : an m-by-n data matrix. 
        Should be already rescaled to non-negative ranges (required by chi2 family) 
        and re-ordered by y. 
    y : class labels
    cmap : color map scheme to use. set None to use default, or set another scheme, e.g., 'gray', 'viridis', 'jet', 'rainbow', etc.
        For small dataset, we recommend 'gray'. For complex dataset, we recommend 'viridis'.
    hyper_param_optimizer : which optimizer to optimize the hyper parameters for each kernel. 
        For real-world dataset, use kes by default. For toy dataset, set None to disable optimizer.
        For multi-class dataset (y has more than 5 classes), use acc.
    scale : whether do feature scaling
    clfs : a list of multi-class classfiers. pass 'all' to use all available classifiers.
    multi_kernels : how many kernels to use for MKL combination. Default is [1].
        If a string, e.g., "poly+gaussian+cosine+spline" or "poly,gaussian,cosine,spline", will use this specific MKL combination.
    multi_kernel_topk : the k best single kernels to use in multi-kernel combinations. Default is 5.
    logplot : whether to output the log-scale plot in parallel.
    plots : whether to ouptut the kernel heatmaps and the scatter plots after PCA / PLS, to check classifiability.
        The PLS tries to maximize the covariance between X and Y.
    embed_title : whether embed the title in the plots. If not, will generate the title in HTML.
    do_cla : whether to do classifiability analysis using the cla package.
    selected_kernel_names : a list of kernel names to be process. 
        If None or 'all', will use all kernels.
    '''

    dic_test_accs = {}
    all_dic_metrics = {}
    html_str = ''

    if scale:
        X = MinMaxScaler().fit_transform(X)
        print('perform min-max scaling')

    if clfs == 'all' or len(clfs) > 0:
        result_html = '<h3>0. no kernel</h3><p>Classification on original dataset</p>'
        _, dic, mc_html = run_multiclass_clfs(X, y, clfs=clfs, show = False)
        dic_test_accs['no kernel'] = dic
        result_html += mc_html
        if output_html:
            html_str += result_html
        if plots:
            IPython.display.display(IPython.display.HTML(result_html))

    # to be safe, perform a re-order
    if (y is not None and len(set(y)) == 2):
        labels = list(set(y))  # re-order X by y
        X = np.vstack((X[y == labels[0]], X[y == labels[1]]))
        y = [labels[0]] * np.sum(y == labels[0]) + \
            [labels[1]] * np.sum(y == labels[1])

    if selected_kernel_names is None or selected_kernel_names == 'all':
        selected_kernel_names = kernel_names

    KX = {} # store the data matrix after each kernel

    for i, key in enumerate(selected_kernel_names):
        best_hparam = None
        best_metric = -np.inf
        title = str(i+1) + '. ' + \
            (kernel_fullnames[key] if key in kernel_fullnames else key)

        if hyper_param_optimizer is not None and key in kernel_hparams:
            for param in kernel_hparams[key]:
                if key in kernel_hparas_divide_n:
                    param = param/X.shape[1]  # divide hparam by data dim
                new_metric = hyper_param_optimizer(
                    X, y, lambda x, y: kernel_dict[key](x, y, param), verbose = verbose)
                if (new_metric > best_metric):
                    best_metric = new_metric
                    best_hparam = param
            title = str(i+1) + '. ' + (kernel_fullnames[key] if key in kernel_fullnames else key) \
                + ('(' + format(best_hparam, '.2g') +
                   ')') if best_hparam is not None else ''

        if hyper_param_optimizer is None or key not in kernel_hparams:
            kns = kernel_dict[key](X, X)
            # metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y), verbose = verbose)
        else:
            kns = kernel_dict[key](X, X, best_hparam)
            # metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y, param), verbose = verbose)

        kns = np.nan_to_num(kns)
        KX[key] = kns

        ######## plot after kernel transforms ########
        
        if plots or output_html:
            _, ax = plt.subplots(1, 2, figsize=(round(len(set(y))/2.0) + 6, round(len(set(y))/4.0) + 3)) # figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3)
            ax[0].imshow(kns, cmap=cmap)
            ax[0].set_axis_off()
            if logplot:
                ax[1].imshow(1+np.log(kns), cmap=cmap)
                ax[1].set_axis_off()

            plt.axis('off')
            if embed_title:
                plt.title(title + '\n' + kernel_formulas[key] + '\n') # + "NMD = %.3g" % metric_nmd)
            else:
                # print(title)
                kernel_title = '<h3>' + title + '</h3>' + '<p>' + kernel_formulas[key].replace('<', '&lt;').replace('>', '&gt;') + '</p>'
                if output_html:
                    html_str += kernel_title
                else:
                    IPython.display.display(IPython.display.HTML(kernel_title)) # <p>' + "NMD = %.3g" % metric_nmd + '</p>'
            if output_html:
                html_str += plt2html(plt)
            plt.show()

            if len(set(y)) <= 3: # disable PCA and LDA plots for more than 3 classes

                ######## scatter plot after PCA ########            
                pca = PCA(n_components=2)  # keep the first 2 components
                X_pca = pca.fit_transform(kns)
                X_pca = np.nan_to_num(X_pca)
                plot_components_2d(X_pca, y)
                plt.title('PCA')

                if output_html:
                    html_str += plt2html(plt)
                plt.show()

                ######## scatter plot after LDA ########

                k_train, k_test, y_train, y_test = train_test_split(
                        kns, y, test_size=0.3, random_state = 2, stratify=y)

                try:
                    lda = LinearDiscriminantAnalysis()
                    X_lda = lda.fit(k_train, y_train).transform(k_test)
                    X_lda = np.nan_to_num(X_lda)

                    lda.score(k_test, y_test)
                    if (X_lda.shape[1] == 1):
                        X_lda = np.hstack((X_lda, np.zeros((X_lda.shape[0], 1))))
                    plot_components_2d(X_lda, y_test)
                    # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
                    plt.title(
                        'LDA (test acc = ' + str(np.round(lda.score(k_test, y_test), 3)) + ')')
                    
                    if output_html:
                        html_str += plt2html(plt)
                    plt.show()

                except Exception as e:
                    if verbose:
                        print('Exception : ', e)
                    # print('X_pls = ', X_pls)
                    # plt.title('PLS')
                    html_str += '<p>' + str(e) + '</p>'

            ######## scatter plot after PLS ########

            if len(set(y)) == 2: # skip PLS for multi-class cases

                try:
                    ''' # using CV
                    kns = np.nan_to_num(kns)
                    pls = PLSRegression(n_components=2, scale=False)

                    k_train, k_test, y_train, y_test = train_test_split(
                        kns, y, test_size=0.3)

                    X_pls = pls.fit(k_train, y_train).transform(k_test)
                    X_pls = np.nan_to_num(X_pls)

                    pls.score(k_test, y_test)
                    plot_components_2d(X_pls, y_test, legends=['C1', 'C2'])
                    # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
                    plt.title(
                        'PLS (R2 = ' + str(np.round(pls.score(k_test, y_test), 3)) + ')')
                    plt.show()
                    '''

                    kns = np.nan_to_num(kns)
                    pls = PLSRegression(n_components=2, scale=False)
                    X_pls = pls.fit(kns, y).transform(kns)
                    X_pls = np.nan_to_num(X_pls)

                    pls.score(kns, y)
                    plot_components_2d(X_pls, y)
                    # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
                    plt.title(
                        'PLS (R2 = ' + str(np.round(pls.score(kns, y), 3)) + ')')
                    if output_html:
                        html_str += plt2html(plt)
                    plt.show()
                except Exception as e:
                    if verbose:
                        print('Exception : ', e)
                    # print('X_pls = ', X_pls)
                    # plt.title('PLS')
                    html_str += '<p>' + str(e) + '</p>'

        ###### Classifiction after kernel transformation #######
        if clfs == 'all' or len(clfs) > 0:
            # result_html = '<h3>classification</h3>'
            _, dic, result_html = run_multiclass_clfs(kns, y, clfs=clfs, show = plots)
            dic_test_accs[key] = dic
            if output_html:
                html_str += result_html

        ###### cla metrics ######
        if do_cla: # disable for now, keep just for reproduce historical results
            kns = np.nan_to_num(np.hstack((kns, np.array(y).reshape(-1, 1))),   # do nan filtering simultaneously for X and y
                                nan=0, posinf=kns.max(), neginf=kns.min())
            _, dic_metrics = get_metrics(kns[:, :-1], kns[:, -1].flatten(), verbose = verbose)
            # dic_metrics['NMD'] = metric_nmd
            all_dic_metrics[key] = dic_metrics

    pickle.dump((KX, dic_test_accs, y), open('single_kernels.pkl', 'wb'))

    if len(multi_kernels) > 0: # multi-kernel cases

        # find top k kernels
        best_KX = {}
        best_top1s = {}
        for k,v in dic_test_accs.items():
            if k == 'no kernel' or not bool(v):
                continue
            
            best_top1 = 0
            for kk, vv in v.items():
                if best_top1 < vv[0]: # top-1 test acc
                    best_top1 = vv[0]
            best_top1s[k] = best_top1

        keys = list(best_top1s.keys())
        values = list(best_top1s.values())
        sorted_value_index = np.argsort(values)[::-1]
        sorted_dict = {keys[i]: values[i] for i in sorted_value_index}

        print('Sorted single-kernel test accs: ', sorted_dict)

        for k in sorted_dict: # sorted(best_top1s.items(),  key = lambda kv:(kv[1], kv[0]), reverse=True):
            best_KX[k] = KX[k]
            if multi_kernel_topk != -1 and len(best_KX) >= multi_kernel_topk:
                break

        if isinstance(multi_kernels, str): # process a specific MKL combination
            ks = multi_kernels.replace("+",",").split(sep=",") # seperator can be + or ,
            return classify_with_mkl(KX, y, ks, clfs=clfs, cmap=cmap, plots=plots, output_html=output_html)

        for multi_kernel in multi_kernels:

            # if multi_kernel == 1: # single kernel case, skip
            #     continue

            for idx, ks in enumerate(combinations(best_KX, multi_kernel)):
                combined = np.zeros((len(y),0))
                mk_title = ''
                for iidx, k in enumerate(ks):
                    if iidx == 0:
                        mk_title += k
                    else:
                        mk_title += ' + ' + k
                    combined = np.hstack((combined, KX[k]))

                result_html = '<h3>' + str(multi_kernel) + '-kernel ' + str(idx+1) + '. ' + mk_title + '</h3>'                
                
                if plots or output_html:
                    IPython.display.display(IPython.display.HTML(result_html)) 
                    plt.figure(figsize=(round(len(set(y))/4.0)*len(ks) + 3, round(len(set(y))/4.0) + 3)) # figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3)
                    plt.imshow(combined, cmap=cmap)
                    plt.axis('off')
                    if output_html:
                        html_str += plt2html(plt)
                    plt.show()
                
                _, dic, result_html_mc = run_multiclass_clfs(combined, y, clfs=clfs, show = plots)
                result_html += result_html_mc
                dic_test_accs[mk_title] = dic
                if output_html:
                    html_str += result_html
                
    return KX, dic_test_accs, all_dic_metrics, html_str

def classify_with_mkl(KX, y, ks, clfs=['LinearDiscriminantAnalysis()'], cmap=None, plots=True, output_html=True):

    combined = np.zeros((len(y),0))
    mk_title = ''
    for iidx, k in enumerate(ks):
        if iidx == 0:
            mk_title += k
        else:
            mk_title += ' + ' + k
        combined = np.hstack((combined, KX[k]))

    result_html = '<h3>' + mk_title + '</h3>'
    
    if plots or output_html:
        IPython.display.display(IPython.display.HTML(result_html)) 
        plt.figure(figsize=(round(len(set(y))/4.0)*len(ks) + 3, round(len(set(y))/4.0) + 3)) # figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3)
        plt.imshow(combined, cmap=cmap)
        plt.axis('off')
        if output_html:
            result_html += plt2html(plt)
        plt.show()
    
    _, dic, result_html_mc = run_multiclass_clfs(combined, y, clfs=clfs, show = plots)
    result_html += result_html_mc
    if output_html:
        return result_html

def visualize_acc_dict(dic_accs):
    '''
    Visualize the acc result dictionary returned by classify_with_kernels()

    Return
    ------
    The best top-1 acc

    Example
    -------
    _, test_acc_dic, _ = classify_with_kernels(X, y, ...)
    visualize_metric_dicts(test_acc_dic) # use in jupyter notebook
    '''

    top1_accs = []
    html_str = '<table>'
    html_str += '<tr>' + '<th>kernel</th>'  + '<th>classifier</th>' + '<th>top-1 acc</th>' + '<th>top-3 acc</th>' + '<th>top-5 acc</th></tr>'
    for k, v in dic_accs.items():
        for kk, vv in v.items():
            accs = [str(round(x,3)) for x in vv]
            top1_accs.append(vv[0])
            html_str += '<tr>' + '<td>' + k + '</td>'  + '<td>' + kk + '</td>' + '<td>' + accs[0] + \
            ('</td>' + '<td>' + accs[1] + '</td>' + '<td>' + accs[2] if len(accs)>2 else '</td><td></td><td>') + '</td></tr>'
            
    html_str += '<tr>' + '<td>best</td>' + '<td></td>' + '<td>' + \
    str(round(np.max(top1_accs),3)) + '</td></tr>'

    html_str += '</table>'
    IPython.display.display(IPython.display.HTML(html_str))

    return round(np.max(top1_accs),3)

def visualize_cla_dicts(dics, plot=True):
    '''
    Example
    -------
    _, _, dics, _ = classify_with_kernels(X, y, cla = True)
    visualize_metric_dicts(dicts) # use in jupyter notebook
    '''

    row_names = []
    column_names = []

    html_str = '<table><tr><th>metric</th>'

    # use the 1st loop to get row and col names
    for kernel in dics:
        column_names.append(kernel)
        html_str += '<th>' + str(kernel) + '</th>'
        for key in dics[kernel]:
            if key not in row_names:
                row_names.append(key)

    html_str += '<th>best</th><th>chart</th>'
    html_str += '</tr>'

    # use the 2nd loop to fill in data
    for row in row_names:
        if row not in metric_polarity_dict:
            continue

        html_str += '<tr><td>' + row + '</td>'
        metrics = []
        for col in column_names:
            metrics.append(dics[col][row] if row in dics[col] else np.nan)
            html_str += '<td>' + \
                (str(round(dics[col][row], 3))
                 if row in dics[col] else '') + '</td>'

        metrics = np.nan_to_num(metrics, nan=np.nan,
                                posinf=np.nan, neginf=np.nan)

        best_metric_value = None
        try:
            if row == 'NMD':  # this is a kernel-specific metric, not listed in metric_polarity_dict
                best_metric_idx = np.nanargmax(metrics)
            else:
                best_metric_idx = metric_polarity_dict[row](metrics)

            # we may have multiple best values
            best_metric_value = metrics[best_metric_idx]
            best_metric_idxs = np.where(metrics == best_metric_value)
        except Exception as e:
            print(e)  # All-NaN slice encountered
            best_metric_idxs = []

        best_kernel_names = str(np.array(column_names)[best_metric_idxs])
        html_str += '<td>' + best_kernel_names + '</td>'
        
        if len(column_names) > 1: # only need to draw bar chart when more than two kernels are compared

            plt.figure(figsize=(20, 3))
            plt.bar(column_names, metrics, alpha=0.7,
                    width=0.6, edgecolor='black', color='white', label=row)
            plt.bar(np.array(column_names)[best_metric_idxs], metrics[best_metric_idxs], alpha=0.7,
                    width=0.6, edgecolor='black', color='gold',
                    label="best kernels: " + best_kernel_names)
            plt.xticks(rotation=40)
            plt.title(row + "\nbest kernels: " + best_kernel_names +
                    "\nbest value: " + str(best_metric_value))
            plt.legend()

            if plot:
                plt.show()
            else:
                html_str += '<td style="min-width:400px">' + plt2html(plt) + '</td>'
                plt.close()

        html_str += '</tr>'

    html_str += '</table>'
    IPython.display.display(IPython.display.HTML(html_str))

    return html_str


def optimize_kernel_hparam(X, y, key, hparams=[], cmap=None):
    '''
    Paramters
    ---------
    X : need to re-order X by y first
    '''
    if hparams is None or len(hparams) <= 0:
        print('hparams is empty. Try built-in values.')
        if key not in kernel_hparams:
            print('Built-in hparams not found. Exit.')
            return

    for h in hparams:
        title = (kernel_fullnames[key] if key in kernel_fullnames else key) \
            + ('(' + format(h, '.2g') + ')')
        plt.imshow(kernel_dict[key](X, X, h), cmap=cmap)
        metric_str = "ACC = %.3g" % acc(X, y, lambda x,
                                   y: kernel_dict[key](x, y, h))
        plt.axis('off')
        plt.title(title + '\n' + kernel_formulas[key] + '\n' + metric_str)
        plt.show()


def cosine_kernel_response_pattern(n=10, cmap='gray', logplot=False, embed_title=False):
    '''
    Special demo for the cosine kernel with 2D dataset. 
    We use equal-angle-interval input. While for other kernels, we use equal-interval input. 
    '''

    comment = 'Unlike other kernels, we use equal-angle-interval (e.g., $ (i/n)*\pi $) input for the cosine kernel.'
    X = []
    for i in range(n):
        theta = math.pi / n * i
        X.append([math.cos(theta), math.sin(theta)])

    X = np.array(X)

    y = [0]*round(n/2) + [1]*(n-round(n/2))
    y = np.array(y)

    kns = cosine_kernel(X, X)
    _, ax = plt.subplots(1, 2, figsize=(12, 6))
    ax[0].imshow(kns, cmap=cmap)
    ax[0].set_axis_off()
    if logplot:
        ax[1].imshow(1+np.log(kns), cmap=cmap)
        ax[1].set_axis_off()
    plt.axis('off')

    metric_str = "ACC = %.3g" % acc(X, y, lambda x, y: kernel_dict['cosine'](x, y))

    if embed_title:
        plt.title(kernel_fullnames['cosine'] + '\n' + kernel_formulas['cosine']
                  + '\n' + metric_str + '\n' + comment)
    else:
        IPython.display.display(IPython.display.HTML('<h3>' + kernel_fullnames['cosine'] + '</h3>' + '<p>' +
                                                     kernel_formulas['cosine'].replace(
            '<', '&lt;').replace('>', '&gt;') + '</p>'
            + '<p>' + metric_str + '</p>' + '<p>' + comment + '</p>'))

    plt.show()


def time_cost_kernels(X, repeat=10, display=True):
    '''
    Time cost of kernels

    return
    ------
    dct_mu : a dict of mean time cost of each kernel
    '''
    dct = {}
    for key, f in kernel_dict.items():
        dct[key] = timeit.repeat(lambda: f(X, X), repeat=repeat, number=1)

    x, y = zip(*dct.items())
    y = np.array(y)

    if display:
        plt.figure(figsize=(15, 4))
        plt.scatter(x, y.mean(axis=1))
        plt.errorbar(x, y.mean(axis=1), y.std(axis=1) * 1,
                     color="dodgerblue", linewidth=1, elinewidth=10, ecolor='r',
                     alpha=0.5, label=' $\mu ± 1 \sigma$ (' + str(y.shape[1]) + ' runs)')  # X.std(axis = 0)

        plt.legend()
        # plt.bar(x, y)
        plt.xticks(rotation=45)
        plt.ylabel('second')
        plt.show()

    dct_mu = {}
    for k, v in zip(x, y.mean(axis=1)):
        dct_mu[k] = v

    return dct_mu


def plot_components_2d(X, y, labels=None, use_markers=False, legends=None, tags=None):
    '''
    Copied from qsi.vis.plot_components_2d, to avoid package dependency.
    '''
    if X.shape[1] < 2:
        print(
            'ERROR: X MUST HAVE AT LEAST 2 FEATURES/COLUMNS! SKIPPING plot_components_2d().')
        return

    if y is None or len(y) == 0:
        labels = [0]  # only one class
    if labels is None:
        labels = set(y)

    _, ax = plt.subplots(figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3))

    # Gray shades can be given as a string encoding a float in the 0-1 range
    colors = ['0.9', '0.1', 'red', 'blue', 'green',
              'orange', 'cyan', 'purple', 'gray']
    if len(labels) > len(colors):
        for i in range(len(labels) - len(colors) + 5):
            colors.append('C' + str(i))

    markers = ['o', 's', '^', 'D', 'H', 'o', 's', '^', 'D',
               'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H']
    if len(labels) > len(markers):
        for i in range(len(labels) - len(markers) + 5):
            markers.append(np.random.choice(markers))

    i = 0

    for label in labels:
        if y is None or len(y) == 0:
            cluster = X
        else:
            cluster = X[np.where(y == label)]
        # print(cluster.shape)

        if use_markers:
            ax.scatter([cluster[:, 0]], [cluster[:, 1]],
                       s=40,
                       marker=markers[i],
                       facecolors='none',
                       edgecolors=colors[i+3],
                       label=(str(legends[i]) if legends is not None else ("Y = " + str(label) + ' (' + str(len(cluster)) + ')')))
        else:
            ax.scatter([cluster[:, 0]], [cluster[:, 1]],
                       s=70,
                       facecolors=colors[i],
                       label=(str(legends[i]) if legends is not None else (
                           "Y = " + str(label) + ' (' + str(len(cluster)) + ')')),
                       edgecolors='black',
                       alpha=.4)  # cmap='tab20'
        i = i+1

    if tags is not None:
        for j, tag in enumerate(tags):
            ax.annotate(str(tag), (X[j, 0] + 0.1, X[j, 1] - 0.1))

    ax.legend()

    ax.axes.xaxis.set_visible(False)
    ax.axes.yaxis.set_visible(False)

    return ax



import os
import sys
import uuid
from flask import Flask, render_template, request
from flaskwebgui import FlaskUI
from qsi import io


if __package__:
    from .. import metrics
else:
    ROOT_DIR = os.path.dirname(os.path.dirname(__file__))
    if ROOT_DIR not in sys.path:
        sys.path.append(ROOT_DIR)
    import metrics

app = Flask(__name__)
app.config['MAX_CONTENT_LENGTH'] = 50 * 1024 * 1024  # limit to 50MB, to avoid 413 error.

@app.route("/", methods=['GET', 'POST'])
def index():
    return render_template("home.html")


@app.route("/submit", methods=['POST'])
def analyze():
    r = ''
    kernel_type = request.form['kernel_type']
    use_sample = request.form["use_sample"]

    if kernel_type == 'all':
        kernel_type = list(metrics.kernel_dict.keys())
    else:
        kernel_type = [kernel_type]

    if (use_sample == 'true' or use_sample is True):
        # path = "ackl/gui/static/sample.csv"
        X, y, X_names, desc, labels = io.load_dataset('salt', x_range=list(range(400, 1000)), display=False)        
    else:
        fn = os.path.dirname(os.path.realpath(
            __file__)) + "/" + str(uuid.uuid4()) + ".csv"
        request.files['dataFile'].save(fn)

        if os.path.isfile(fn) is False:
            r = 'File ' + fn + ' does not exist.'
            return {'message': 'success', 'html': r}

        X, y, X_names, labels = io.open_dataset(fn)
        desc = fn

    try:
        r += '<h6>Load dataset</h6>'
        if len(X.shape) == 2:

            if y is None:
                r += io.draw_average(X, X_names, output_html=True)
            else:
                r += io.draw_class_average(X, y, X_names, labels, output_html=True)

            r += io.scatter_plot(X, y, labels=labels, output_html=True)

        r += '<p>' + desc + '</p><hr/>'
        r += '<h6>Kernel Transformation</h6>'

        _, dic_test_accs, dic, s = metrics.classify_with_kernels(X, y, scale = True,
                                        do_cla=True,
                                        clfs='all', # ['LinearDiscriminantAnalysis()'],
                                        plots=True,
                                        logplot=True,
                                        output_html= True,
                                        selected_kernel_names = kernel_type)

        r += s
        r += metrics.visualize_metric_dicts(dic, plot=False)

    except Exception as e:
        r += '<b>' + str(e) + '</b>'

    return {'message': 'success', 'html': r}

if __name__ == '__main__':
    FlaskUI(app=app, server="flask", port=5052).run()
